open Client

(* A few manually constructed terms. *)

let x =
  ML.Var "x"

let y =
  ML.Var "y"

let id =
  ML.Abs ("x", x)

let delta =
  ML.Abs ("x", ML.App (x, x))

let deltadelta =
  ML.App (delta, delta)

let idid =
  ML.App (id, id)

let k =
  ML.Abs ("x", ML.Abs ("y", x))

let genid =
  ML.Let ("x", id, x)

let genidid =
  ML.Let ("x", id, ML.App (x, x))

let genkidid =
  ML.Let ("x", ML.App (k, id), ML.App (x, id))

let genkidid2 =
  ML.Let ("x", ML.App (ML.App (k, id), id), x)

let app_pair = (* ill-typed *)
  ML.App (ML.Tuple [id; id], id)

let abs_match_with =
  ML.(
    Abs(
      "x",
      Match(
        Tuple[],
        [ (PTuple[], Tuple[]) ]
      )
    )
  )

(* option *)
let option_env =

  (* type 'a option = None | Some of 'a *)
  let option_typedecl =
    let open ML in
    let open Datatype in {
    name = Type "option";
    type_params = [ "a" ];
    data_kind = Variant_type;
    labels_descr = [
      {
        label_name = Label"None";
        type_name = Type "option";
        arg_type = TyProduct [];
      } ; {
        label_name = Label "Some";
        type_name = Type "option";
        arg_type = TyVar "a";
      }
    ]
  } in

  Env.add_typedecl Env.empty option_typedecl

let none = ML.Variant ( Datatype.Label "None" , ML.Tuple [] )

let some =
  ML.Variant (
    Datatype.Label "Some",
    id
  )

let match_none = ML.(
    Match (none, [
      PVariant (Datatype.Label "None", PTuple []), none ;
      PVariant (Datatype.Label "Some", PTuple ([PVar "x"])), x ;
    ])
  )

let match_some = ML.(
    Match (some, [
      PVariant (Datatype.Label "None", PTuple []), none ;
      PVariant (Datatype.Label "Some", PWildcard), none
    ])
  )

(* list *)
let list_env =

  (* type 'a list = Nil | Cons of 'a * 'a list *)
  let list_typedecl =
    let open ML in
    let open Datatype in {
    name = Type "list";
    type_params = [ "a" ];
    data_kind = Variant_type;
    labels_descr = [
      {
        label_name = Label "Nil";
        type_name = Type "list";
        arg_type = TyProduct [];
      } ; {
        label_name = Label "Cons";
        type_name = Type "list";
        arg_type = TyProduct [ TyVar "a" ; TyConstr (Type "list", [ TyVar "a" ]) ];
      }
    ]
  } in

  Env.add_typedecl option_env list_typedecl

let nil = ML.Variant ( Datatype.Label "Nil" , ML.Tuple [] )

let list =
  ML.Variant (
    Datatype.Label "Cons",
    ML.Tuple [ id ; nil ]
  )

(* tree *)
let tree_env =

  (* type 'a tree = Leaf | Node of 'a tree * 'a * 'a tree *)
  let tree_typedecl =
    let open ML in
    let open Datatype in {
    name = Type "tree";
    type_params = [ "a" ];
    data_kind = Variant_type;
    labels_descr = [
      {
        label_name = Label "Leaf";
        type_name = Type "tree";
        arg_type = TyProduct []
      } ; {
        label_name = Label "Node";
        type_name = Type "tree";
        arg_type = TyProduct [
          TyConstr (Type "tree", [ TyVar "a" ]);
          TyVar "a";
          TyConstr (Type "tree", [ TyVar "a" ]);
        ]
      }
    ];
  } in

  Env.add_typedecl list_env tree_typedecl

let leaf = ML.Variant ( Datatype.Label "Leaf" , ML.Tuple [] )

let node =
  ML.Variant (
    Datatype.Label "Node",
    ML.Tuple [
      leaf ;
      id ;
      leaf ;
    ]
  )

let () =
  assert Test.(Log.with_log CheckML.test Env.empty idid);
  assert Test.(Log.with_log CheckML.test Env.empty genid);
  assert Test.(Log.with_log CheckML.test Env.empty genidid);
  assert Test.(Log.with_log CheckML.test Env.empty genkidid);
  assert Test.(Log.with_log CheckML.test Env.empty genkidid2);
  assert Test.(Log.with_log CheckML.test option_env none);
  assert Test.(Log.with_log CheckML.test option_env some);
  assert Test.(Log.with_log CheckML.test list_env nil);
  assert Test.(Log.with_log CheckML.test list_env list);
  assert Test.(Log.with_log CheckML.test tree_env leaf);
  assert Test.(Log.with_log CheckML.test tree_env node);
  assert Test.(Log.with_log CheckML.test Env.empty abs_match_with);
  assert Test.(Log.with_log CheckML.test option_env match_none);
  assert Test.(Log.with_log CheckML.test option_env match_some);
  (* we include some printing below because Dune
     appears to only show tests that show some output. *)
  print_endline "TestML: all tests passed."
